#ifndef __internal_h__
#define __internal_h__

#include "jfaud.h"

#if defined __GNUC__
typedef long long int64;
#elif defined _MSC_VER
typedef __int64 int64;
#endif

#ifdef linux
// shouldn't need a calling convention explicitly given...
#define __cdecl
#endif

#ifndef max
#define max(a,b) (((a)<(b))?(b):(a))
#endif
#ifndef min
#define min(a,b) (((a)<(b))?(a):(b))
#endif

#ifdef __GNUC__
#define PACKED __attribute__ ((packed))
#define UNUSED __attribute__ ((unused))
#define inline __inline__
#else
#define PACKED
#define UNUSED
#define __attribute__(x)
#define inline _inline
#ifdef _MSC_VER
#define snprintf _snprintf
#define strncasecmp strnicmp
#define strcasecmp stricmp
#endif
#endif

#define swaplong(a)  ((((a)&0xff000000)>>24)|(((a)&0xff0000)>>8)|(((a)&0xff00)<<8)|(((a)&0xff)<<24))
#define swapshort(a) ((((a)&0xff00)>>8)|(((a)&0xff)<<8))

#ifdef BIG_ENDIAN
#define LittleLong(a) swaplong((a))
#define BigLong(a) (a)
#define LittleShort(a) swapshort((a))
#define BigShort(a) (a)
#define MustSwapBig (0)
#define MustSwapLittle (1)

#define ExtractBigLong(p) (*(unsigned long *)p)
static inline unsigned long ExtractLittleLong(const unsigned char *p)
{
	unsigned long l = 0;
	l  = (unsigned long)p[0] << 24;
	l |= (unsigned long)p[1] << 16;
	l |= (unsigned long)p[2] << 8;
	l |= (unsigned long)p[3];
	return l;
}
#define ExtractBigShort(p) (*(unsigned short *)p)
static inline unsigned short ExtractLittleShort(const unsigned char *p)
{
	unsigned short l = 0;
	l |= (unsigned short)p[0] << 8;
	l |= (unsigned short)p[1];
	return l;
}

#else

#define LittleLong(a) (a)
#define BigLong(a) swaplong((a))
#define LittleShort(a) (a)
#define BigShort(a) swapshort((a))
#define MustSwapBig (1)
#define MustSwapLittle (0)

static inline unsigned long ExtractBigLong(const unsigned char *p)
{
	unsigned long l = 0;
	l  = (unsigned long)p[0] << 24;
	l |= (unsigned long)p[1] << 16;
	l |= (unsigned long)p[2] << 8;
	l |= (unsigned long)p[3];
	return l;
}
#define ExtractLittleLong(p) (*(unsigned long *)p)
static inline unsigned short ExtractBigShort(const unsigned char *p)
{
	unsigned short l = 0;
	l |= (unsigned short)p[0] << 8;
	l |= (unsigned short)p[1];
	return l;
}
#define ExtractLittleShort(p) (*(unsigned short *)p)
#endif

typedef union {
	struct {
		unsigned int flags;
		unsigned int ident;
	} a;
	int64 ident;
} fileident;

#define fileidentflag_doublespeed	(1)
#define fileidentflag_fxnearest		(2)
#define fileidentflag_fxstertomono	(4)

#define fileidentfmt "{%d:%x}"
#define fileidentval(x) x a.flags, x a.ident

#include "buffer.h"
#include "soundfile.h"
#include "channel.h"
#include "alout.h"
#include "channel.h"
#include "fx.h"
#include "thread.h"
#include "midiout.h"
#include "midiseq.h"
#include "cda.h"

extern JFAudCfg conf;	// the current configuration settings

char *jsprintf(const char *fmt, ...) __attribute__((format(printf, 1, 2)));
void logmsg(const char *s);
#ifndef NODEBUGLOG
#ifdef __GNUC__
#define debuglog(l,s) debugmsg(l,s,__FILE__,__PRETTY_FUNCTION__,__LINE__)
#else
#define debuglog(l,s) debugmsg(l,s,__FILE__,NULL,__LINE__)
#endif
#else
#define debuglog(l,s)
#endif
void debugmsg(int level, const char *s, const char *fn, const char *func, int line);

fileident hashfn(const char *fn, const char *subfn);

static inline JFAudProp mkpropv(int prop, float a, float b, float c) {
	JFAudProp p;
	p.prop = prop; p.val.v[0] = a; p.val.v[1] = b; p.val.v[2] = c;
	return p;
}

#endif // __internal_h__
