#ifndef __soundfile_h__
#define __soundfile_h__

typedef struct {
	JFAudFH fh;
	fileident ident;
	int readeridx;
	char *filename;

	struct SoundFile_t *sf;
} File;

int sysf_open(const char *fn, const char *subfn, JFAudFH *h, JFAudRawFormat **raw);
int sysf_close(JFAudFH *h);
int sysf_read(JFAudFH *h, void *buf, int len);
int sysf_seek(JFAudFH *h, int off, int whence);
int sysf_tell(JFAudFH *h);
int sysf_filelen(JFAudFH *h);



typedef enum {
	type_wave,
	type_midi,
	type_cda
} filetype;

typedef enum {
	fmt_raw,	// this is a special case
	fmt_riffwav,
	fmt_voc,
	fmt_au,
	fmt_vorbis,
	fmt_mpeg,
	fmt_flac,
	fmt_mod
} waveformformat;

#define SoundFile_defn \
	filetype type;
typedef struct SoundFile_t {
	SoundFile_defn
} SoundFile;

#define WaveformFile_defn \
	SoundFile_defn \
	waveformformat format;
typedef struct {
	WaveformFile_defn
} WaveformFile;

#define MidiFile_defn \
	SoundFile_defn
typedef struct {
	MidiFile_defn
} MidiFile;

#define SoundFileReader_defn \
	const char *name; \
	jfauderr (*init)(void); \
	jfauderr (*uninit)(void); \
	\
	filetype type; \
	jfauderr (*test)(File *); \
	jfauderr (*unload)(File *, char *noclose); \
	jfauderr (*getmetainfo)(File *, JFAudMetaInfo what, char *buf, int *len);
typedef struct {
	SoundFileReader_defn
} SoundFileReader;

typedef struct {
	SoundFileReader_defn

	waveformformat format;
	jfauderr (*readsamples)(File *, int n, PcmBuffer *buf, char loop);
	jfauderr (*readallsamples)(File *, PcmBuffer *buf);
	jfauderr (*getplaytime)(File *, float *len);
} WaveformFileReader;

typedef struct {
	SoundFileReader_defn

	jfauderr (*getdata)(File *, MidiBuffer *buf);
} MidiFileReader;

jfauderr filereaders_init(void);
jfauderr filereaders_uninit(void);
jfauderr filereaders_identify(File *file);
jfauderr filereaders_unload(File *file, char *noclose);

jfauderr filereaders_wave_readsamples(File *, int n, PcmBuffer *buf, char loop);
jfauderr filereaders_wave_readallsamples(File *, PcmBuffer *buf);
jfauderr filereaders_wave_getplaytime(File *, float *len);

jfauderr filereaders_midi_getdata(File *, MidiBuffer *buf);

jfauderr soundfile_open(File **fh, const char *fn, const char *subfn);
jfauderr soundfile_close(File *fh);

#endif // __soundfile_h__
