#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h>
#include "jfaud.h"

#ifdef _WIN32
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <io.h>
#include <fcntl.h>
#else
#include <unistd.h>
#include <fcntl.h>
#ifndef O_BINARY
#define O_BINARY 0
#endif
#endif

#ifndef min
#define min(x,y) ((x)<(y)?(x):(y))
#endif

#define testsnd "testsnd/koi-44.ogg"
//#define testsnd "testsnd/sin440-96s2.wav"

typedef struct {
	int realstart, reallen, realpos;
	JFAudRawFormat fmt;
} Kwv;

int testopen(const char *fn, const char *subfn, JFAudFH *h, JFAudRawFormat **raw)
{
	Kwv kwv;
	const char *f;
	*raw = NULL;
	memset(h,0,sizeof(JFAudFH));
	h->fh = open(fn, O_RDONLY|O_BINARY);
	if (h->fh < 0) { return h->fh; }
	f = fn; while (*f) f++;

	if (tolower(*(--f)) != 'v' || tolower(*(--f)) != 'w' || tolower(*(--f)) != 'k' || *(--f) != '.')
		{ return h->fh; }

	// the file extension suggests a KWV file, so try and find the subfn in it
	if (!subfn) { return h->fh; }

	kwv.realstart = 0;
	{
		int i,j, numwaves, wavleng, repstart, repleng, finetune;
		char instname[16], found = 0;

		if (read(h->fh, &i, 4) != 4) goto error;
		if (i != 0) goto error;
		if (read(h->fh, &numwaves, 4) != 4) goto error;
		kwv.realstart = numwaves * (16+4*4);
		for (i=0;i<numwaves;i++) {
			if (read(h->fh, instname, 16) != 16) goto error;
			if (read(h->fh, &wavleng, 4) != 4) goto error;
			if (read(h->fh, &repstart, 4) != 4) goto error;
			if (read(h->fh, &repleng, 4) != 4) goto error;
			if (read(h->fh, &finetune, 4) != 4) goto error;
			for (j=0;j<16 && subfn[j];j++) {
				if (tolower(subfn[j]) != tolower(instname[j])) {
					found = 0;
					break;
				} else found = 1;
			}
			if (found) break; else kwv.realstart += wavleng;
		}
		if (!found) goto error;
		kwv.reallen = wavleng;
		lseek(h->fh, kwv.realstart, SEEK_SET);
	}

	kwv.realpos = 0;
	kwv.fmt.samplerate = 11025;
	kwv.fmt.channels   = 1;
	kwv.fmt.bitspersample = 8;
	kwv.fmt.bigendian = 0;

	h->user = (void*)malloc(sizeof(kwv));
	if (!h->user) {
		close(h->fh);
		h->fh = -1;
		return -1;
	}
	memcpy(h->user, &kwv, sizeof(kwv));
	*raw = &((Kwv*)h->user)->fmt;
	return h->fh;

error:
	lseek(h->fh,0,SEEK_SET);
	return h->fh;
}
int testclose(JFAudFH *h)
{
	if (h->user) free(h->user);
	if (h->fh >= 0) return close(h->fh);
	return -1;
}
int testread(JFAudFH *h, void *buf, int len)
{
	int r;
	Kwv *kwv = (Kwv *)h->user;
	if (kwv) len = min( kwv->reallen - (kwv->realpos - kwv->realstart), len);
	if (len < 0) len = 0;
	r = read(h->fh, buf, len);
	if (kwv) kwv->realpos += len;
	return r;
}
int testseek(JFAudFH *h, int off, int whence)
{
	int s;
	Kwv *kwv = (Kwv *)h->user;
	if (kwv) {
		switch (whence) {
			case SEEK_CUR:
				s = lseek(h->fh, off, SEEK_CUR);
				break;
			case SEEK_SET:
				off = min(off, kwv->reallen);
				s = lseek(h->fh, kwv->realstart + off, SEEK_SET);
				break;
			case SEEK_END:
				if (off > 0) off = 0;
				s = lseek(h->fh, kwv->realstart + kwv->reallen + off, SEEK_SET);
				break;
			default: return -1;
		}
		s -= kwv->realstart;
		kwv->realpos = s;
		return s;
	}
	return lseek(h->fh, off, whence);
}
int testtell(JFAudFH *h)
{
	Kwv *kwv = (Kwv *)h->user;
	if (kwv) return kwv->realpos;
	return lseek(h->fh,0,SEEK_CUR);
}
int testfilelen(JFAudFH *h)
{
	Kwv *kwv = (Kwv *)h->user;
#if defined _MSC_VER || defined __WATCOMC__
	if (kwv) return kwv->reallen;
	return filelength(h->fh);
#else
	int len, cur;
	if (kwv) return kwv->reallen;
	cur = lseek(h->fh,0,SEEK_CUR);
	len = lseek(h->fh,0,SEEK_END);
	lseek(h->fh,cur,SEEK_SET);
	return len;
#endif
}

int main(int argc, char **argv)
{
	jfauderr err;
	JFAudCfg cfg = {
		"", "", "",
		4,22050,1,
		testopen,testclose,testread,testseek,testtell,testfilelen,NULL
	};
	JFAudProp pr[5];
	const char *sndfile = testsnd, *subfile = NULL;

	int hnd;
	float t=0.0, step = 2.0*3.14159/90.0;

	if (argc>=2) sndfile = argv[1];
	if (argc>=3) subfile = argv[2];

	pr[0].prop = JFAudProp_loop;
	pr[0].val.i = 1;
	pr[1].prop = JFAudProp_fx;
	pr[1].val.i = 2;	// 2=stereo-to-mono, 1=nearest
	pr[2].prop = JFAudProp_position;
	pr[2].val.v[0] = 1.0; pr[2].val.v[1] = 0.0; pr[2].val.v[2] = 0.0;
	pr[3].prop = JFAudProp_posrel;
	pr[3].val.i = 1;
	pr[4].prop = JFAudProp_rolloff;
	pr[4].val.f = 1.0;
	
	err = JFAud_init(&cfg);
	printf("JFAud_init() returned status: %s\n", JFAud_errstr(err));
	if (err != jfauderr_noerror) return 1;

	if (cfg.wavedev && (err = JFAud_isinited(JFAudInit_wave)) != jfauderr_noerror) {
		printf("Wave device failed to initialise: %s\n", JFAud_errstr(err));
		//JFAud_uninit();
		//return 1;
	}

	if (cfg.mididev && (err = JFAud_isinited(JFAudInit_midi)) != jfauderr_noerror) {
		printf("MIDI device failed to initialise: %s\n", JFAud_errstr(err));
		//JFAud_uninit();
		//return 1;
	}
	
	JFAud_setdebuglevel(DEBUGLVL_ALL);
	//JFAud_setstreamthresh(20.f);

	err = JFAud_playmusic(/*"testsnd/fishpolk.mid",NULL*/"*CDA","1",JFAudPlayMode_playing,0,NULL);
	if (err != jfauderr_noerror) printf("Failed to play music: %s\n", JFAud_errstr(err));

	err = JFAud_playsound(&hnd,sndfile,subfile,1,JFAudPlayMode_playing,sizeof(pr)/sizeof(pr[0]),&pr[0]);
	printf("JFAud_playsound() returned status: %s\n", JFAud_errstr(err));
	if (err == jfauderr_noerror) {
		do {
			pr[0].prop = JFAudProp_position;
			pr[0].val.v[0] = 1.0*sin(t);
			pr[0].val.v[1] = 1.0*cos(t);
			pr[0].val.v[2] = 0.0;

			t += step;
			
			JFAud_update();
			JFAud_setsoundprops(hnd, 1, pr);
#ifdef _WIN32
			Sleep(100);	// 100 millisec
#else
			usleep(100000);
#endif
		} while (t < 2.0*3.14159*3.0);
	}
	
	err = JFAud_uninit();
	printf("JFAud_uninit() returned status: %s\n", JFAud_errstr(err));
	if (err != jfauderr_noerror) return 1;

	return 0;
}
